from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast, overload

import numpy as np

from contourpy import FillType, LineType, max_threads

if TYPE_CHECKING:
    import numpy.typing as npt

    import contourpy._contourpy as cpy


point_dtype = np.float64
code_dtype = np.uint8
offset_dtype = np.uint32


def all_class_names() -> list[str]:
    return [
        "Mpl2005ContourGenerator",
        "Mpl2014ContourGenerator",
        "SerialContourGenerator",
        "ThreadedContourGenerator",
    ]


def all_names(exclude: str | None = None) -> list[str]:
    all_ = ["mpl2005", "mpl2014", "serial", "threaded"]
    if exclude is not None:
        all_.remove(exclude)
    return all_


def all_names_and_fill_types() -> list[tuple[str, FillType]]:
    return [
        ("mpl2005", FillType.OuterCode),
        ("mpl2014", FillType.OuterCode),
        ("serial", FillType.OuterCode),
        ("serial", FillType.OuterOffset),
        ("serial", FillType.ChunkCombinedCode),
        ("serial", FillType.ChunkCombinedOffset),
        ("serial", FillType.ChunkCombinedCodeOffset),
        ("serial", FillType.ChunkCombinedOffsetOffset),
        ("threaded", FillType.OuterCode),
        ("threaded", FillType.OuterOffset),
        ("threaded", FillType.ChunkCombinedCode),
        ("threaded", FillType.ChunkCombinedOffset),
        ("threaded", FillType.ChunkCombinedCodeOffset),
        ("threaded", FillType.ChunkCombinedOffsetOffset),
    ]


def all_names_and_line_types() -> list[tuple[str, LineType]]:
    return [
        ("mpl2005", LineType.SeparateCode),
        ("mpl2014", LineType.SeparateCode),
        ("serial", LineType.Separate),
        ("serial", LineType.SeparateCode),
        ("serial", LineType.ChunkCombinedCode),
        ("serial", LineType.ChunkCombinedOffset),
        ("threaded", LineType.Separate),
        ("threaded", LineType.SeparateCode),
        ("threaded", LineType.ChunkCombinedCode),
        ("threaded", LineType.ChunkCombinedOffset),
    ]


def corner_mask_names() -> list[str]:
    return ["mpl2014", "serial", "threaded"]


def quad_as_tri_names() -> list[str]:
    return ["serial", "threaded"]


def all_fill_types_str_value() -> list[tuple[str, int]]:
    return [
        ("OuterCode", 201),
        ("OuterOffset", 202),
        ("ChunkCombinedCode", 203),
        ("ChunkCombinedOffset", 204),
        ("ChunkCombinedCodeOffset", 205),
        ("ChunkCombinedOffsetOffset", 206),
    ]


def all_line_types_str_value() -> list[tuple[str, int]]:
    return [
        ("Separate", 101),
        ("SeparateCode", 102),
        ("ChunkCombinedCode", 103),
        ("ChunkCombinedOffset", 104),
    ]


def all_z_interps_str_value() -> list[tuple[str, int]]:
    return [
        ("Linear", 1),
        ("Log", 2),
    ]


def thread_counts() -> list[int]:
    thread_counts = [2, 3]
    return list(filter(lambda n: n <= max(max_threads(), 1), thread_counts))


def assert_point_array(points: cpy.PointArray) -> int:
    assert isinstance(points, np.ndarray)
    assert points.dtype == point_dtype
    assert points.ndim == 2
    assert points.shape[1] == 2
    npoints = points.shape[0]
    assert npoints >= 1
    return npoints


def assert_code_array(codes: cpy.CodeArray, npoints: int) -> None:
    assert isinstance(codes, np.ndarray)
    assert codes.dtype == code_dtype
    assert codes.ndim == 1
    assert len(codes) == npoints
    assert codes[0] == 1


def assert_offset_array(offsets: cpy.OffsetArray, max_offset: int) -> int:
    assert isinstance(offsets, np.ndarray)
    assert offsets.dtype == offset_dtype
    assert offsets.ndim == 1
    assert len(offsets) > 1
    assert offsets[0] == 0
    assert offsets[-1] == max_offset
    assert np.all(np.diff(offsets) > 0)
    return len(offsets)


def assert_filled(filled: cpy.FillReturn, fill_type: FillType) -> None:
    if fill_type == FillType.OuterCode:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_OuterCode, filled)
        assert isinstance(filled, tuple) and len(filled) == 2
        polygons, codes = filled
        assert isinstance(polygons, list)
        assert isinstance(codes, list)
        assert len(polygons) == len(codes)
        for polygon, code in zip(polygons, codes):
            npoints = assert_point_array(polygon)
            assert_code_array(code, npoints)
    elif fill_type == FillType.OuterOffset:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_OuterOffset, filled)
        assert isinstance(filled, tuple) and len(filled) == 2
        polygons, offsets = filled
        assert isinstance(polygons, list)
        assert isinstance(offsets, list)
        assert len(polygons) == len(offsets)
        for polygon, offset in zip(polygons, offsets):
            npoints = assert_point_array(polygon)
            assert_offset_array(offset, npoints)
    elif fill_type == FillType.ChunkCombinedCode:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedCode, filled)
        assert isinstance(filled, tuple) and len(filled) == 2
        chunk_polygons, chunk_codes = filled
        assert isinstance(chunk_polygons, list)
        assert isinstance(chunk_codes, list)
        assert len(chunk_polygons) == len(chunk_codes)
        for polygons_or_none, codes_or_none in zip(chunk_polygons, chunk_codes):
            if polygons_or_none is None:
                assert codes_or_none is None
            else:
                if TYPE_CHECKING:
                    assert codes_or_none is not None
                npoints = assert_point_array(polygons_or_none)
                assert_code_array(codes_or_none, npoints)
    elif fill_type == FillType.ChunkCombinedOffset:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedOffset, filled)
        assert isinstance(filled, tuple) and len(filled) == 2
        chunk_polygons, chunk_offsets = filled
        assert isinstance(chunk_polygons, list)
        assert isinstance(chunk_offsets, list)
        assert len(chunk_polygons) == len(chunk_offsets)
        for polygons_or_none, offsets_or_none in zip(chunk_polygons, chunk_offsets):
            if polygons_or_none is None:
                assert offsets_or_none is None
            else:
                if TYPE_CHECKING:
                    assert offsets_or_none is not None
                npoints = assert_point_array(polygons_or_none)
                assert_offset_array(offsets_or_none, npoints)
    elif fill_type == FillType.ChunkCombinedCodeOffset:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedCodeOffset, filled)
        assert isinstance(filled, tuple) and len(filled) == 3
        chunk_polygons, chunk_codes, chunk_outer_offsets = filled
        assert isinstance(chunk_polygons, list)
        assert isinstance(chunk_codes, list)
        assert isinstance(chunk_outer_offsets, list)
        assert len(chunk_polygons) == len(chunk_codes) == len(chunk_outer_offsets)
        for polygons_or_none, codes_or_none, outer_offsets_or_none in zip(
                chunk_polygons, chunk_codes, chunk_outer_offsets):
            if polygons_or_none is None:
                assert codes_or_none is None and outer_offsets_or_none is None
            else:
                if TYPE_CHECKING:
                    assert codes_or_none is not None and outer_offsets_or_none is not None
                npoints = assert_point_array(polygons_or_none)
                assert_code_array(codes_or_none, npoints)
                assert_offset_array(outer_offsets_or_none, npoints)
    elif fill_type == FillType.ChunkCombinedOffsetOffset:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedOffsetOffset, filled)
        assert isinstance(filled, tuple) and len(filled) == 3
        chunk_polygons, chunk_offsets, chunk_outer_offsets = filled
        assert isinstance(chunk_polygons, list)
        assert isinstance(chunk_offsets, list)
        assert isinstance(chunk_outer_offsets, list)
        assert len(chunk_polygons) == len(chunk_offsets) == len(chunk_outer_offsets)
        for polygons_or_none, offsets_or_none, outer_offsets_or_none in zip(
                chunk_polygons, chunk_offsets, chunk_outer_offsets):
            if polygons_or_none is None:
                assert offsets_or_none is None and outer_offsets_or_none is None
            else:
                if TYPE_CHECKING:
                    assert offsets_or_none is not None and outer_offsets_or_none is not None
                npoints = assert_point_array(polygons_or_none)
                noffsets = assert_offset_array(offsets_or_none, npoints)
                assert_offset_array(outer_offsets_or_none, noffsets-1)
    else:
        raise RuntimeError(f"Unexpected fill_type {fill_type}")


def assert_lines(lines: cpy.LineReturn, line_type: LineType) -> None:
    if line_type == LineType.Separate:
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_Separate, lines)
        assert isinstance(lines, list)
        for line in lines:
            assert_point_array(line)
    elif line_type == LineType.SeparateCode:
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_SeparateCode, lines)
        assert isinstance(lines, tuple) and len(lines) == 2
        lines, codes = lines
        assert isinstance(lines, list)
        assert isinstance(codes, list)
        assert len(lines) == len(codes)
        for line, code in zip(lines, codes):
            npoints = assert_point_array(line)
            assert_code_array(code, npoints)
    elif line_type == LineType.ChunkCombinedCode:
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_ChunkCombinedCode, lines)
        assert isinstance(lines, tuple) and len(lines) == 2
        chunk_lines, chunk_codes = lines
        assert isinstance(chunk_lines, list)
        assert isinstance(chunk_codes, list)
        assert len(chunk_lines) == len(chunk_codes)
        for lines_or_none, codes_or_none in zip(chunk_lines, chunk_codes):
            if lines_or_none is None:
                assert codes_or_none is None
            else:
                if TYPE_CHECKING:
                    assert codes_or_none is not None
                npoints = assert_point_array(lines_or_none)
                assert_code_array(codes_or_none, npoints)
    elif line_type == LineType.ChunkCombinedOffset:
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_ChunkCombinedOffset, lines)
        assert isinstance(lines, tuple) and len(lines) == 2
        chunk_lines, chunk_offsets = lines
        assert isinstance(chunk_lines, list)
        assert isinstance(chunk_offsets, list)
        assert len(chunk_lines) == len(chunk_offsets)
        for lines_or_none, offsets_or_none in zip(chunk_lines, chunk_offsets):
            if lines_or_none is None:
                assert offsets_or_none is None
            else:
                if TYPE_CHECKING:
                    assert offsets_or_none is not None
                npoints = assert_point_array(lines_or_none)
                assert_offset_array(offsets_or_none, npoints)
    else:
        raise RuntimeError(f"Unexpected line_type {line_type}")


@overload
def sort_by_first_xy(lines: list[cpy.PointArray]) -> list[cpy.PointArray]:
    ...


@overload
def sort_by_first_xy(
    lines: list[cpy.PointArray], others: list[npt.NDArray[Any]],
) -> tuple[list[cpy.PointArray], list[npt.NDArray[Any]]]:
    ...


def sort_by_first_xy(
    lines: list[cpy.PointArray], others: list[npt.NDArray[Any]] | None = None,
) -> list[cpy.PointArray] | tuple[list[cpy.PointArray], list[npt.NDArray[Any]]]:
    first_xy = np.array([line[0] for line in lines])
    order = np.lexsort((first_xy[:, 1], first_xy[:, 0]))
    lines = [lines[o] for o in order]
    if others is None:
        return lines
    else:
        others = [others[o] for o in order]
        return lines, others



from __future__ import annotations

from copy import deepcopy
from datetime import datetime
from typing import Any

from asv.benchmark import Benchmark
from asv.benchmarks import Benchmarks
from asv.config import Config
from asv.results import Results, iter_results_for_machine
from asv.statistics import get_err

from contourpy import FillType, LineType


class Loader:
    _config: Config
    _benchmarks: Benchmarks
    _machine: str
    _results: Results

    def __init__(self, machine: str | None = None) -> None:
        self._config = Config.load()
        self._benchmarks = Benchmarks.load(self._config)

        if machine is None:
            import platform
            machine = platform.uname()[1]

        latest_results = None
        for results in iter_results_for_machine(self._config.results_dir, machine):
            if latest_results is None or results.date > latest_results.date:
                latest_results = results
        if latest_results is None:
            raise RuntimeError("No results found for machine {machine}")
        print(latest_results.commit_hash, datetime.fromtimestamp(latest_results.date/1000.0))

        self._results = latest_results
        self._machine = machine

    def _find_benchmark_by_name(self, name: str) -> Benchmark:
        for k, v in self._benchmarks.items():
            if k.endswith(name):
                return v
        raise RuntimeError(f"Cannot find benchmark with name {name}")

    @property
    def commit(self) -> str:
        return self._results.commit_hash  # type: ignore[no-any-return]

    def get(self, benchmark_name: str, **kwargs: Any) -> dict[str, Any]:
        benchmark = self._find_benchmark_by_name(benchmark_name)
        param_names = benchmark["param_names"]
        params = deepcopy(benchmark["params"])
        for name, value in kwargs.items():
            index = param_names.index(name)
            if isinstance(value, list):
                params[index] = [repr(item) for item in value]
            else:
                params[index] = [repr(value)]

        stats = self._results.get_result_stats(benchmark["name"], params)
        values = self._results.get_result_value(benchmark["name"], params)

        ret = {}
        for name, param in zip(param_names, params):
            for i, item in enumerate(param):
                if isinstance(item, str):
                    if item[0] == "'" and item[-1] == "'":
                        item = item[1:-1]

                    if item.startswith("<FillType"):
                        item = FillType(int(item[item.index(" "):-1]))
                    elif item.startswith("<LineType"):
                        item = LineType(int(item[item.index(" "):-1]))
                    elif item == "False":
                        item = False
                    elif item == "True":
                        item = True
                    else:
                        try:
                            item = int(item)
                        except ValueError:
                            pass
                    param[i] = item
            ret[name] = param[0] if len(param) == 1 else param

        if values[0] is None:
            ret["mean"] = ret["error"] = None
        else:
            ret["mean"] = values
            ret["error"] = [get_err(v, s) for v, s in zip(values, stats)]

        return ret

    @property
    def machine(self) -> str:
        return self._machine



from __future__ import annotations

import os
from shutil import copyfile
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    import io


def compare_images(
    test_buffer: io.BytesIO,
    baseline_filename: str,
    test_filename_suffix: str | None = None,
    max_threshold: int | None = None,
    mean_threshold: float | None = None,
) -> None:
    from PIL import Image

    if max_threshold is None:
        max_threshold = 100
    if mean_threshold is None:
        mean_threshold = 0.02

    if test_filename_suffix:
        basename, extension = os.path.splitext(baseline_filename)
        test_filename = f"{basename}_{test_filename_suffix}{extension}"
    else:
        test_filename = baseline_filename

    max_diff = None
    mean_diff = None

    try:
        test_image = np.asarray(Image.open(test_buffer).convert("RGB"))
        test_image = test_image.astype(np.int16)

        full_baseline_filename = os.path.join("tests", "baseline_images", baseline_filename)
        assert os.path.isfile(full_baseline_filename)

        baseline_image = np.asarray(Image.open(full_baseline_filename).convert("RGB"))
        baseline_image = baseline_image.astype(np.int16)

        diff = np.abs(test_image - baseline_image)
        if np.dtype(np.intp).itemsize == 4:
            # 32-bit Matplotlib agg backend RGB components can be 1 out compared to 64-bit due to
            # slightly different rounding.  So ignore differences of 1 on 32-bit.
            diff[diff == 1] = 0
        max_diff = diff.max()
        mean_diff = diff.mean()

        assert max_diff < max_threshold and mean_diff < mean_threshold
    except AssertionError:
        # Write test image.
        result_directory = "result_images"
        if not os.path.exists(result_directory):
            os.makedirs(result_directory)

        full_test_filename = os.path.join(result_directory, test_filename)
        with open(full_test_filename, "wb") as f:
            f.write(test_buffer.getbuffer())

        if max_diff is not None and mean_diff is not None:
            basename, extension = os.path.splitext(test_filename)

            # Copy expected file.
            expected_filename = f"{basename}-expected{extension}"
            expected_filename = os.path.join(result_directory, expected_filename)
            copyfile(full_baseline_filename, expected_filename)

            print(f"diffs: max {max_diff}, mean {mean_diff:.4f} "
                  f"(thresholds max {max_threshold}, mean {mean_threshold})")

            # Write difference image.
            difference_filename = f"{basename}-diff{extension}"
            difference_filename = os.path.join(result_directory, difference_filename)

            diff = diff.astype(np.float64)*10
            diff = np.clip(diff, 0, 255).astype(np.uint8)
            Image.fromarray(diff).save(difference_filename, format="png")

        raise



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from asv.util import human_value
from loader import Loader
import matplotlib.pyplot as plt
import numpy as np

from contourpy import FillType, LineType

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from matplotlib.text import Annotation


# Default fill/line types that exist in all algorithms.
default_fill_type = FillType.OuterCode
default_line_type = LineType.SeparateCode


def capital_letters_to_newlines(text: str) -> str:
    return re.sub(r"([a-z0-9])([A-Z])", r"\1\n\2", text)


def get_corner_mask_label(corner_mask: bool | str) -> str:
    if corner_mask == "no mask":
        return "no mask"
    else:
        return f"corner_mask={corner_mask}"


def get_style(name: str, corner_mask: bool | str) -> tuple[str, str, str, float]:
    # Colors from Paul Tol's colorblind friendly light scheme (https://personal.sron.nl/~pault)
    colors = {
        "mpl2005": "#eedd88",   # light yellow.
        "mpl2014": "#ee8866",   # orange.
        "serial": "#77aadd",    # light blue.
        "threaded": "#99ddff",  # light cyan.
    }

    hatches = {
        "no mask": "",
        False: "---",
        True: "///",
    }

    edge_color = "#222222"

    return colors[name], edge_color, hatches[corner_mask], 0.5


def with_time_units(value: float, error: float | None = None) -> str:
    # ASV's human_value() doesn't put a space between numbers and units.
    # See e.g. https://physics.nist.gov/cuu/Units/checklist.html
    with_units = human_value(value, "seconds", error)
    return re.sub(r"(?<=\S)([a-zA-Z]+)$", r" \1", with_units)


def by_name_and_type(loader: Loader, filled: bool, dataset: str, render: bool, n: int) -> None:
    show_error = False
    corner_masks: list[str | bool] = ["no mask", False, True]
    filled_str = "filled" if filled else "lines"
    title = f"{filled_str} {dataset} n={n} {'(calculate and render)' if render else ''}"

    nbars = 3
    width = 1.0 / (nbars + 1)
    ntypes = len(FillType.__members__) if filled else len(LineType.__members__)

    for mode in ["light", "dark"]:
        plt.style.use("default" if mode == "light" else "dark_background")

        fig, ax = plt.subplots(figsize=(8.5, 6))
        xticklabels = []

        for name in ["mpl2005", "mpl2014", "serial"]:
            bname = "serial" if name == "serial" else "mpl20xx"
            benchmarks_name = f"time_{filled_str}_{bname}{'_render' if render else ''}"

            if name == "serial":
                xs = 2 + np.arange(ntypes)
            else:
                xs = np.array(0 if name == "mpl2005" else 1)

            i = 0
            for corner_mask in corner_masks:
                kwargs = dict(name=name, dataset=dataset, corner_mask=corner_mask, n=n)

                results = loader.get(benchmarks_name, **kwargs)
                if results["name"] != name:
                    raise RuntimeError("Loader returning wrong name: {name} != {results['name']}")

                if results["mean"] is None:
                    continue

                name = results["name"]
                mean = results["mean"]
                error = results["error"]
                if corner_mask == "no mask":
                    types = results["fill_type" if filled else "line_type"]
                    if not isinstance(types, list):
                        types = [types]
                    xticklabels += [name + str(t).split(".")[1] for t in types]

                color, edge_color, hatch, line_width = get_style(name, corner_mask)
                offset = width*(i - 0.5*(nbars - 1))
                label = f"{name} {get_corner_mask_label(corner_mask)}"
                yerr = error if show_error else None
                mean = np.asarray(mean, dtype=np.float64)  # None -> nan.

                rects = ax.bar(
                    xs + offset, mean, width, yerr=yerr, color=color, edgecolor=edge_color,
                    hatch=hatch, linewidth=line_width, capsize=4, label=label, zorder=3)
                if show_error:
                    labels = [with_time_units(m, s) for m, s in zip(mean, error)]
                else:
                    labels = [with_time_units(m) for m in mean]
                ax.bar_label(rects, labels, padding=5, rotation="vertical", size="medium")

                i += 1

        if filled and not render:
            if dataset == "random":
                ax.set_ylim(0, 2.6)
            else:
                ax.set_ylim(0, 0.3)
        elif not filled and render and dataset == "simple":
            ax.set_ylim(0, 0.4)
        else:
            ax.set_ylim(0, ax.get_ylim()[1]*1.1)  # Magic number.

        loc: str | tuple[float, float] = "best"
        if not filled and render and dataset == "random":
            loc = "lower left"
        elif render and dataset == "simple":
            loc = "lower right"
        elif filled and render and dataset == "random":
            loc = (0.51, 0.6)
        ax.legend(loc=loc, framealpha=0.9)

        ax.grid(axis="y", c="k" if mode == "light" else "w", alpha=0.2)
        ax.set_xticks(np.arange(ntypes+2))
        xticklabels = list(map(capital_letters_to_newlines, xticklabels))
        ax.set_xticklabels(xticklabels)
        ax.set_ylabel("Time (seconds)")
        ax.set_title(title)
        for spine in ax.spines.values():
            spine.set_zorder(5)
        fig.tight_layout()

        filename = f"{filled_str}_{dataset}_{n}{'_render' if render else ''}_{mode}.svg"
        print(f"Saving {filename}")
        fig.savefig(filename, transparent=True)


def comparison_two_benchmarks(
    loader: Loader, filled: bool, dataset: str, varying: str, varying_values: list[float],
) -> None:
    if varying == "thread_count":
        file_prefix = "threaded"
    elif varying == "total_chunk_count":
        file_prefix = "chunk"
    else:
        raise RuntimeError(f"Invalid varying field '{varying}'")

    show_error = False
    show_speedups = (varying == "thread_count")
    n = 1000
    corner_mask = "no mask"

    filled_str = "filled" if filled else "lines"
    kwargs = dict(dataset=dataset, corner_mask=corner_mask, n=n)
    if varying == "thread_count":
        kwargs["total_chunk_count"] = 40

    name0 = "serial"
    name1 = "threaded" if varying == "thread_count" else "serial"

    kwargs["name"] = name0
    if varying == "thread_count":
        benchmarks_name = f"time_{filled_str}_{name0}_chunk"
    else:
        benchmarks_name = f"time_{filled_str}_{name0}"
    results = loader.get(benchmarks_name, **kwargs)
    fill_or_line_type = results["fill_type"] if filled else results["line_type"]
    ntype = len(fill_or_line_type)
    mean0 = results["mean"]
    error0 = results["error"]

    kwargs["name"] = name1
    kwargs[varying] = varying_values
    if varying == "thread_count":
        benchmarks_name = f"time_{filled_str}_{name1}"
    else:
        benchmarks_name = f"time_{filled_str}_{name1}_chunk"
    results = loader.get(benchmarks_name, **kwargs)
    mean1 = results["mean"]
    error1 = results["error"]

    varying_count = len(varying_values)
    xs = np.arange(ntype*(varying_count+2))
    xs.shape = (ntype, varying_count+2)

    speedups = np.expand_dims(mean0, axis=1) / np.reshape(mean1, (ntype, varying_count))
    speedups = speedups.ravel()

    def in_bar_label(ax: Axes, rect: Annotation, value: str) -> None:
        kwargs = dict(fontsize="medium", ha="center", va="bottom", color="k")
        if varying != "thread_count":
            kwargs["rotation"] = "vertical"
        ax.annotate(value, (rect.xy[0] + 0.5*rect.get_width(), rect.xy[1]), **kwargs)

    for mode in ["light", "dark"]:
        plt.style.use("default" if mode == "light" else "dark_background")
        fig, ax = plt.subplots(figsize=(8.5, 6))

        # Serial bars.
        color, edge_color, hatch, line_width = get_style(name0, corner_mask)
        if varying == "thread_count":
            label = f"{name0} {get_corner_mask_label(corner_mask)}"
        else:
            label = None
        rects = ax.bar(xs[:, 0], mean0, width=1, color=color, edgecolor=edge_color, hatch=hatch,
                       linewidth=line_width, label=label, zorder=3)
        if show_error:
            labels = [with_time_units(m, s) for m, s in zip(mean0, error0)]
        else:
            labels = [with_time_units(m) for m in mean0]
        ax.bar_label(rects, labels, padding=5, rotation="vertical", size="medium")
        if varying != "thread_count":
            for rect in rects:
                in_bar_label(ax, rect, " 1")

        # Threaded bars.
        color, edge_color, hatch, line_width = get_style(name1, corner_mask)
        label = varying.replace("_", " ")
        label = f"{name1} {get_corner_mask_label(corner_mask)}\n({label} shown at bottom of bar)"
        rects = ax.bar(xs[:, 1:-1].ravel(), mean1, width=1, color=color, edgecolor=edge_color,
                       hatch=hatch, linewidth=line_width, label=label, zorder=3)
        labels = []
        for i, (mean, error, speedup) in enumerate(zip(mean1, error1, speedups)):
            if show_error:
                label = with_time_units(mean, error)
            else:
                label = with_time_units(mean)
            if show_speedups and i % varying_count > 0:
                label += f" (x {speedup:.2f})"
            labels.append(label)
        ax.bar_label(rects, labels, padding=5, rotation="vertical", size="medium")
        for rect, value in zip(rects, np.tile(varying_values, ntype)):
            in_bar_label(ax, rect, f" {value}")

        if dataset == "random":
            ymax = 2.0 if filled else 1.4
        elif varying == "thread_count":
            ymax = ax.get_ylim()[1]*1.32
        else:
            ymax = ax.get_ylim()[1]*1.25
        ax.set_ylim(0, ymax)

        ax.set_xticks(xs[:, 0] + 0.5*varying_count)
        xticklabels = [str(t).split(".")[1] for t in fill_or_line_type]
        xticklabels = list(map(capital_letters_to_newlines, xticklabels))
        ax.set_xticklabels(xticklabels)

        ax.legend(loc="upper right", framealpha=0.9)
        ax.grid(axis="y", c="k" if mode == "light" else "w", alpha=0.2)
        ax.set_ylabel("Time (seconds)")
        ax.set_title(f"{filled_str} {dataset} n={n}")
        fig.tight_layout()

        filename = f"{file_prefix}_{filled_str}_{dataset}_{mode}.svg"
        print(f"Saving {filename}")
        fig.savefig(filename, transparent=True)


def main() -> None:
    loader = Loader()

    print(f"Saving benchmark plots for machine={loader.machine} commit={loader.commit[:7]}")

    for filled in [False, True]:
        for dataset in ["random", "simple"]:
            for render in [False, True]:
                by_name_and_type(loader, filled, dataset, render, 1000)

    for filled in [False, True]:
        for dataset in ["random", "simple"]:
            comparison_two_benchmarks(loader, filled, dataset, "total_chunk_count",
                                      [4, 12, 40, 120])

    for filled in [False, True]:
        for dataset in ["random", "simple"]:
            comparison_two_benchmarks(loader, filled, dataset, "thread_count", [1, 2, 4, 6])


if __name__ == "__main__":
    main()



from __future__ import annotations

import pytest

from contourpy.util import build_config


@pytest.fixture
def all_keys() -> list[str]:
    return [
        "b_ndebug",
        "b_vscrt",
        "build_cpu_endian",
        "build_cpu_family",
        "build_cpu_system",
        "build_cpu",
        "build_dir",
        "build_options",
        "buildtype",
        "compile_command",
        "compiler_name",
        "compiler_version",
        "contourpy_version",
        "cpp_std",
        "cross_build",
        "debug",
        "host_cpu_endian",
        "host_cpu_family",
        "host_cpu_system",
        "host_cpu",
        "linker_id",
        "meson_backend",
        "meson_version",
        "mesonpy_version",
        "optimization",
        "pybind11_version",
        "python_install_dir",
        "python_path",
        "python_version",
        "source_dir",
        "vsenv",
    ]


def test_build_config(all_keys: list[str]) -> None:
    config = build_config()

    for key in all_keys:
        # key exists and its value is a non-empty string
        value = config.pop(key)
        assert isinstance(value, str) and len(value) > 0

    # assert no keys left
    assert len(config) == 0



from __future__ import annotations

from abc import abstractmethod
from enum import Enum
import io
from typing import TYPE_CHECKING, Any, Literal, cast

import matplotlib.pyplot as plt
import numpy as np

from contourpy import FillType, LineType, contour_generator

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure
    import numpy.typing as npt

    from contourpy._contourpy import (
        CoordinateArray, FillReturn_OuterCode, LineReturn_SeparateCode, MaskArray,
    )


class PointType(Enum):
    CORNER = 0
    LOWER = 1
    UPPER = 2


class Corner(Enum):
    NW = 1
    NE = 0
    SW = 3
    SE = 2


class Config:
    corner_mask: bool
    name: str
    quad_as_tri: bool
    show_text: bool

    arrowsize: float
    axes: npt.NDArray[Axes]  # Set in derived classes.
    axes_index: int
    fill_alpha: float
    fig: Figure  # Set in derived classes.
    fontsize: float
    gap: float
    grid_kwargs: dict[str, str | float]
    mask: MaskArray | Literal[False]
    marker_size: float
    text_gap: float
    title_fontsize: float
    x: CoordinateArray
    y: CoordinateArray

    def __init__(
        self, name: str, corner_mask: bool, quad_as_tri: bool, show_text: bool,
    ) -> None:
        self.name = name
        self.corner_mask = corner_mask
        self.quad_as_tri = quad_as_tri
        self.show_text = show_text

        self.arrowsize = 0.1
        self.fontsize = 8
        self.title_fontsize = 8
        self.marker_size = 3.5
        self.gap = 0.1
        self.text_gap = self.gap/2
        self.grid_kwargs = dict(color="gray", linewidth=0.35)
        self.fill_alpha = 0.2

        self.x, self.y = np.meshgrid([0.0, 1.0], [0.0, 1.0])
        self.mask = False

        import matplotlib
        matplotlib.use("Agg")

    def __del__(self) -> None:
        if self.fig:
            plt.close(self.fig)

    def _arrow(
        self, ax: Axes, line_start: CoordinateArray, line_end: CoordinateArray, color: str,
    ) -> None:
        mid = 0.5*(line_start + line_end)
        along = line_end - line_start
        along /= np.sqrt(np.dot(along, along))  # Unit vector.
        right = np.asarray((along[1], -along[0]))
        arrow = np.stack((
            mid - (along*0.5 - right)*self.arrowsize,
            mid + along*0.5*self.arrowsize,
            mid - (along*0.5 + right)*self.arrowsize))
        ax.plot(arrow[:, 0], arrow[:, 1], "-", c=color)

    @abstractmethod
    def _decode_config(self, config: int, corner: Corner | None = None) -> tuple[int | None, ...]:
        pass

    @abstractmethod
    def _quad_lines(
        self,
        ax: Axes,
        z: np.ma.MaskedArray[Any, Any],
        zlower: float,
        zupper: float,
        corner: Corner | None,
    ) -> None:
        pass

    def _next_quad(
        self,
        config: int,
        corner: Corner | None = None,
        suffix: str = "",
        set_0: float = 0,
        set_1: float = 1,
        set_2: float = 2,
    ) -> None:
        zlower = 0.5
        zupper = 1.5

        ax = self.axes[self.axes_index]
        nw, ne, sw, se = self._decode_config(config, corner)

        # Quad edge.
        ax.axis("off")
        if corner is None:
            ax.plot([0, 1, 1, 0, 0], [0, 0, 1, 1, 0], **self.grid_kwargs)
            if self.quad_as_tri and not nw == ne == sw == se:
                ax.plot([0, 1], [0, 1], [0, 1], [1, 0], **self.grid_kwargs)
            title = f"{nw}{ne}{sw}{se}={config}{suffix}"
        elif corner == Corner.NW:
            ax.plot([0, 1, 0, 0], [0, 1, 1, 0], **self.grid_kwargs)
            title = f"{nw}{ne}{sw}={config}{suffix}"
        elif corner == Corner.NE:
            ax.plot([1, 1, 0, 1], [0, 1, 1, 0], **self.grid_kwargs)
            title = f"{nw}{ne}{se}={config}{suffix}"
        elif corner == Corner.SW:
            ax.plot([0, 1, 0, 0], [0, 0, 1, 0], **self.grid_kwargs)
            title = f"{nw}{sw}{se}={config}{suffix}"
        elif corner == Corner.SE:
            ax.plot([0, 1, 1, 0], [0, 0, 1, 0], **self.grid_kwargs)
            title = f"{ne}{sw}{se}={config}{suffix}"
        else:
            raise RuntimeError("Invalid corner")

        if self.show_text:
            ax.set_title(title, size=self.title_fontsize)

        # Axes bounds include gap for arrows and corner points.
        ax.set_xlim(-self.gap, 1.0+self.gap)
        ax.set_ylim(-self.gap, 1.0+self.gap)

        if self.show_text:
            # Text for corner and optional middle z-levels.
            fontsize = self.fontsize
            if corner != Corner.NW:
                ax.text(1+self.text_gap, 0, se, va="center", size=fontsize, ha="left")
            if corner != Corner.SW:
                ax.text(1+self.text_gap, 1, ne, va="center", size=fontsize, ha="left")
            if corner != Corner.NE:
                ax.text(-self.text_gap, 0, sw, va="center", size=fontsize, ha="right")
            if corner != Corner.SE:
                ax.text(-self.text_gap, 1, nw, va="center", size=fontsize, ha="right")
            if suffix:  # Suffix (without brackets) is the z-level of quad middle.
                ax.text(0.5, 0.5, suffix[1:-1], va="center", size=fontsize, ha="center")

        lookup = {0: set_0, 1: set_1, 2: set_2, None: None}
        z_nw = lookup[nw]
        z_ne = lookup[ne]
        z_sw = lookup[sw]
        z_se = lookup[se]

        z = np.asarray([[z_sw, z_se], [z_nw, z_ne]], dtype=np.float64)
        z_masked = np.ma.array(z, mask=self.mask)  # type: ignore[no-untyped-call]

        if suffix:
            zmean = z.mean()
            if suffix == "(0)":
                if zmean > zlower:
                    raise RuntimeError(
                        f"Inconsistent zlower {zlower} for config {config}{suffix} with middle_z "
                        f"{zmean}")
            elif suffix in ["(1)", "(>0)", "(<2)"]:
                if zmean < zlower:
                    raise RuntimeError(
                        f"Inconsistent zlower {zlower} for config {config}{suffix} with middle_z "
                        f"{zmean}")
                elif zmean > zupper:
                    raise RuntimeError(
                        f"Inconsistent zupper {zupper} for config {config}{suffix} with middle_z "
                        f"{zmean}")
            elif suffix == "(2)":
                if zmean < zupper:
                    raise RuntimeError(
                        f"Inconsistent zupper {zupper} for config {config}{suffix} with middle_z "
                        f"{zmean}")
            else:
                raise RuntimeError(f"Unexpected suffix {suffix} for config {config}")

        self._quad_lines(ax, z_masked, zlower, zupper, corner)
        self.axes_index += 1

    def _set_mask(self, corner: Corner | None) -> None:
        if corner is None:
            self.mask = False
        else:
            self.mask = np.zeros(4, dtype=bool)
            self.mask[corner.value] = True
            self.mask.shape = (2, 2)

    def save_to_buffer(self) -> io.BytesIO:
        buf = io.BytesIO()
        self.fig.savefig(buf, format="png")
        buf.seek(0)
        return buf

    def save_to_file(self, filename: str) -> None:
        self.fig.savefig(filename)


class ConfigFilledCommon(Config):
    def __init__(self, name: str, corner_mask: bool, quad_as_tri: bool, show_text: bool) -> None:
        super().__init__(name, corner_mask, quad_as_tri, show_text)

    def _decode_config(self, config: int, corner: Corner | None = None) -> tuple[int | None, ...]:
        if corner is None:
            nw = (config >> 6) & 0x3
            ne = (config >> 4) & 0x3
            sw = (config >> 2) & 0x3
            se = config & 0x3
        else:
            a = (config >> 4) & 0x3
            b = (config >> 2) & 0x3
            c = config & 0x3
            if corner == Corner.NW:
                [nw, ne, sw, se] = [a, b, c, None]
            elif corner == Corner.NE:
                [nw, ne, sw, se] = [a, b, None, c]
            elif corner == Corner.SW:
                [nw, ne, sw, se] = [a, None, b, c]
            else:  # Corner.SE
                [nw, ne, sw, se] = [None, a, b, c]
        if nw == 3 or ne == 3 or sw == 3 or se == 3:
            raise ValueError("Invalid config")
        return nw, ne, sw, se

    def _quad_lines(
        self,
        ax: Axes,
        z: np.ma.MaskedArray[Any, Any],
        zlower: float,
        zupper: float,
        corner: Corner | None,
    ) -> None:
        cont_gen = contour_generator(
            self.x, self.y, z, name=self.name, corner_mask=self.corner_mask,
            quad_as_tri=self.quad_as_tri, fill_type=FillType.OuterCode)

        filled = cont_gen.filled(zlower, zupper)
        assert cont_gen.fill_type == FillType.OuterCode
        if TYPE_CHECKING:
            filled = cast(FillReturn_OuterCode, filled)
        lines = filled[0]

        # May be 0..2 polygons, and there cannot be any holes.
        for points in lines:
            n = len(points)

            ax.fill(points[:, 0], points[:, 1], c="C2", alpha=self.fill_alpha, ec=None)

            # Classify points, either corner, lower level or upper level.
            point_types = np.empty(n, dtype=PointType)
            for i in range(n-1):
                xy = points[i]
                x_on_quad_edge = (xy[0] in (0.0, 1.0))
                y_on_quad_edge = (xy[1] in (0.0, 1.0))
                if x_on_quad_edge and y_on_quad_edge:
                    point_types[i] = PointType.CORNER
                elif corner is not None or x_on_quad_edge or y_on_quad_edge:
                    if (self._interp(z, xy[0], xy[1], corner) < 0.5*(zlower + zupper)):
                        point_types[i] = PointType.LOWER
                    else:
                        point_types[i] = PointType.UPPER
                else:
                    # Point on quad diagonal from middle point, so same type as previous point.
                    point_types[i] = point_types[i-1]
            # End point is the same as the start point.
            point_types[-1] = point_types[0]

            # Draw lines.
            for i in range(n-1):
                s = points[i]
                e = points[i+1]
                c = "C2"
                if point_types[i] == point_types[i+1] and point_types[i] != PointType.CORNER:
                    c = "C3" if point_types[i] == PointType.LOWER else "C0"
                ax.plot([s[0], e[0]], [s[1], e[1]], "-", c=c)

                # Arrows on boundary and contour levels (the latter only if not quad_as_tri).
                if point_types[i] == point_types[i+1] and (
                        point_types[i] == PointType.CORNER or not self.quad_as_tri):
                    self._arrow(ax, s, e, c)

            # Arrows on contour levels (lower and upper) for quad_as_tri.  Only want a single
            # arrow on each line strip, so find and use the longest line segment.  Care needed
            # identifying the longest due to rounding error, so find all segments within a small
            # length of the longest and take the max index of these.
            if self.quad_as_tri:
                for point_type in [PointType.LOWER, PointType.UPPER]:
                    mask = np.equal(point_types, point_type)  # type: ignore[call-overload]
                    start_indices = np.nonzero(np.logical_and(mask[1:], ~mask[:-1]))[0]
                    for start_index in start_indices:
                        start_index = (start_index + 1) % (n-1)

                        # Roll point_types array so start segment of line strip is at front.
                        rolled = np.roll(point_types[:-1], -start_index)
                        end_offset = np.nonzero(rolled != point_type)[0][0]

                        pts = points[start_index:start_index+end_offset]  # Line strip points.
                        diff = np.diff(pts, axis=0)
                        lengths = np.sum(np.square(diff), axis=1)  # Segment lengths.
                        offset = np.nonzero(lengths > lengths.max() - 1e-6)[0].max()

                        i = (start_index + offset) % (n-1)
                        c = "C3" if point_type == PointType.LOWER else "C0"
                        self._arrow(ax, points[i], points[i+1], c)

            # Draw markers.
            for i in range(n-1):
                if point_types[i] != PointType.CORNER:
                    c = "C3" if point_types[i] == PointType.LOWER else "C0"
                    ax.plot(points[i][0], points[i][1], "o", c=c, ms=self.marker_size)

    def _interp(
        self,
        zquad: CoordinateArray,
        x: CoordinateArray,
        y: CoordinateArray,
        corner: Corner | None = None,
    ) -> float:
        # Interpolate zquad to determine value of z at (x, y).  Could use
        # scipy.interp2d for this, but do not want scipy as a dependency just
        # for this.
        # (x, y) must lie on one of the edges of the quad.
        # zquad[0,0] = sw, zquad[0,1] = se, zquad[1,0] = nw, zquad[1,1] = ne
        ret: float

        if x == 0.0:
            ret = zquad[0, 0]*(1.0-y) + zquad[1, 0]*y
        elif x == 1.0:
            ret = zquad[0, 1]*(1.0-y) + zquad[1, 1]*y
        elif y == 0.0:
            ret = zquad[0, 0]*(1.0-x) + zquad[0, 1]*x
        elif y == 1.0:
            ret = zquad[1, 0]*(1.0-x) + zquad[1, 1]*x
        else:
            # corner is not None
            if corner in (Corner.NW, Corner.SE):
                ret = zquad[0, 0]*(1.0-y) + zquad[1, 1]*y
            else:
                ret = zquad[0, 1]*(1.0-y) + zquad[1, 0]*y

        return ret


class ConfigFilled(ConfigFilledCommon):
    def __init__(self, name: str, quad_as_tri: bool = False, show_text: bool = True) -> None:
        super().__init__(name, False, quad_as_tri, show_text)

        subplot_kw = dict(aspect="equal")
        if self.quad_as_tri:
            self.fig, axes = plt.subplots(18, 12, figsize=(10.4, 17.1), subplot_kw=subplot_kw)
        else:
            self.fig, axes = plt.subplots(9, 11, figsize=(9.4, 8.4), subplot_kw=subplot_kw)
        self.axes = axes.flatten()

        self.axes_index = 0
        for config in range(171):
            try:
                (nw, ne, sw, se) = self._decode_config(config)
            except ValueError:
                continue

            assert nw is not None and ne is not None and sw is not None and se is not None
            all = np.asarray((nw, ne, sw, se))

            # A quad is degenerate if all 4 quad edges include either the lower
            # or upper contour levels.
            degenerate_lower = (se == nw == 0 and sw > 0 and ne > 0) or \
                               (sw == ne == 0 and se > 0 and nw > 0)
            degenerate_upper = (se == nw == 2 and sw < 2 and ne < 2) or \
                               (sw == ne == 2 and se < 2 and nw < 2)

            if self.quad_as_tri:
                if not nw == ne == sw == se:
                    zmax = all.max()
                    zmin = all.min()
                    zsum = all.sum()
                    if zmin == 0 and zmax == 1:
                        lookup_0 = {1: 0, 2: -1, 3: -1.8}
                        self._next_quad(config, suffix="(0)", set_0=lookup_0[zsum])

                        lookup_1 = {1: 0.4, 2: 0.35, 3: 0}
                        self._next_quad(config, suffix="(1)", set_0=lookup_1[zsum])
                    elif zmin == 0 and zmax == 2:
                        count_1 = np.count_nonzero(all == 1)
                        lookup_0 = {2: -0.3, 3: -1, 4: -2-count_1/2, 5: -5, 6: -6}
                        self._next_quad(config, suffix="(0)", set_0=lookup_0[zsum])

                        if zsum <= 3:
                            self._next_quad(config, suffix="(1)", set_2=3.01)
                        elif zsum <= 5:
                            self._next_quad(config, suffix="(1)")
                        elif zsum == 6:
                            self._next_quad(config, suffix="(1)", set_0=-1)

                        lookup_2 = {2: 8.01, 3: 7.01, 4: 4+count_1/2, 5: 3.5, 6: 2.5}
                        self._next_quad(config, suffix="(2)", set_2=lookup_2[zsum])
                    elif zmin == 1 and zmax == 2:
                        lookup_1 = {5: 2, 6: 1.65, 7: 1.6}
                        self._next_quad(config, suffix="(1)", set_2=lookup_1[zsum])

                        lookup_2 = {5: 4, 6: 3, 7: 2}
                        self._next_quad(config, suffix="(2)", set_2=lookup_2[zsum])
                    else:
                        raise RuntimeError(f"Invalid combination of zmin {zmin} and zmax {zmax}")
                else:
                    self._next_quad(config)
            else:  # !quad_as_tri
                if degenerate_lower and degenerate_upper:
                    # middle needs to be 0, 1 and 2.
                    self._next_quad(config, suffix="(0)", set_0=-1.01)
                    self._next_quad(config, suffix="(1)")
                    self._next_quad(config, suffix="(2)", set_2=3.01)
                elif degenerate_lower:
                    # middle needs to be 0 and >0.
                    self._next_quad(config, suffix="(0)", set_0=-0.01 if all.max() == 1 else -0.51)
                    self._next_quad(config, suffix="(>0)", set_1=1.01)
                elif degenerate_upper:
                    # middle needs to be <2 and 2.
                    self._next_quad(config, suffix="(<2)")
                    self._next_quad(config, suffix="(2)", set_2=2.01 if all.min() == 1 else 2.51)
                else:
                    # Not degenerate.
                    self._next_quad(config)

        # Hide unwanted axes.
        for i in range(self.axes_index, len(self.axes)):
            self.axes[i].axis("off")

        self.fig.tight_layout()


class ConfigFilledCorner(ConfigFilledCommon):
    def __init__(self, name: str, show_text: bool = True) -> None:
        super().__init__(name, corner_mask=True, quad_as_tri=False, show_text=show_text)

        self.fig, axes = plt.subplots(8, 14, figsize=(12.1, 7.6), subplot_kw={"aspect": "equal"})
        self.axes = axes.flatten()

        corners = [Corner.NW, Corner.NE, Corner.SW, Corner.SE]
        self.axes_index = 0
        for icorner in range(4):
            corner = corners[icorner]
            self._set_mask(corner)
            for config in range(43):
                try:
                    _ = self._decode_config(config, corner)
                except ValueError:
                    continue

                self._next_quad(config, corner=corner)

            self.axes[self.axes_index].axis("off")
            self.axes_index += 1

        self.fig.tight_layout()


class ConfigLinesCommon(Config):
    def __init__(self, name: str, corner_mask: bool, quad_as_tri: bool, show_text: bool) -> None:
        super().__init__(name, corner_mask, quad_as_tri, show_text)

    def _decode_config(self, config: int, corner: Corner | None = None) -> tuple[int | None, ...]:
        if corner is None:
            nw = (config >> 3) & 0x1
            ne = (config >> 2) & 0x1
            sw = (config >> 1) & 0x1
            se = config & 0x1
        else:
            a = (config >> 2) & 0x1
            b = (config >> 1) & 0x1
            c = config & 0x1
            if corner == Corner.NW:
                [nw, ne, sw, se] = [a, b, c, None]
            elif corner == Corner.NE:
                [nw, ne, sw, se] = [a, b, None, c]
            elif corner == Corner.SW:
                [nw, ne, sw, se] = [a, None, b, c]
            else:  # Corner.SE
                [nw, ne, sw, se] = [None, a, b, c]
        return nw, ne, sw, se

    def _quad_lines(
        self,
        ax: Axes,
        z: np.ma.MaskedArray[Any, Any],
        zlower: float,
        zupper: float,
        corner: Corner | None,
    ) -> None:
        cont_gen = contour_generator(
            self.x, self.y, z, name=self.name, corner_mask=self.corner_mask,
            quad_as_tri=self.quad_as_tri, line_type=LineType.SeparateCode)

        lines_and_codes = cont_gen.lines(zlower)
        assert cont_gen.line_type == LineType.SeparateCode
        if TYPE_CHECKING:
            lines_and_codes = cast(LineReturn_SeparateCode, lines_and_codes)
        lines = lines_and_codes[0]

        for points in lines:
            ax.plot(points[:, 0], points[:, 1], "o-", c="C3", ms=self.marker_size)

            # Single arrow in middle segment (if even number of points) or the following
            # segment (if odd number of points).
            n = len(points)
            i = (n-1) // 2
            self._arrow(ax, points[i], points[i+1], "C3")


class ConfigLines(ConfigLinesCommon):
    def __init__(self, name: str, quad_as_tri: bool = False, show_text: bool = True) -> None:
        super().__init__(name, False, quad_as_tri, show_text)

        subplot_kw = dict(aspect="equal")
        if self.quad_as_tri:
            self.fig, axes = plt.subplots(5, 6, figsize=(5.2, 4.75), subplot_kw=subplot_kw)
        else:
            self.fig, axes = plt.subplots(3, 6, figsize=(5.2, 2.85), subplot_kw=subplot_kw)
        self.axes = axes.flatten()

        self.axes_index = 0
        for config in range(16):
            (nw, ne, sw, se) = self._decode_config(config)

            assert nw is not None and ne is not None and sw is not None and se is not None

            # A quad is degenerate if all 4 quad edges include the z level.
            degenerate = (se == nw == 0 and sw > 0 and ne > 0) or \
                         (sw == ne == 0 and se > 0 and nw > 0)

            if degenerate or (self.quad_as_tri and not nw == ne == sw == se):
                if self.quad_as_tri:
                    count_1 = np.count_nonzero((nw, ne, sw, se))
                    count_0 = 4 - count_1
                    self._next_quad(config, suffix="(0)", set_0=1.01 - count_1)
                    self._next_quad(config, suffix="(1)", set_1=0.01 + count_0)
                else:
                    self._next_quad(config, suffix="(0)", set_1=0.99)
                    self._next_quad(config, suffix="(1)", set_1=1.01)
            else:
                self._next_quad(config)

        self.fig.tight_layout()


class ConfigLinesCorner(ConfigLinesCommon):
    # All 4 corners plotted together.
    def __init__(self, name: str, show_text: bool = True) -> None:
        super().__init__(name, corner_mask=True, quad_as_tri=False, show_text=show_text)

        self.fig, axes = plt.subplots(4, 8, figsize=(6.93, 3.8), subplot_kw={"aspect": "equal"})
        self.axes = axes.flatten()

        self.axes_index = 0
        corners = [Corner.NW, Corner.NE, Corner.SW, Corner.SE]
        for icorner in range(4):
            corner = corners[icorner]
            self._set_mask(corner)

            for config in range(8):
                self._next_quad(config, corner)

        self.fig.tight_layout()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, cast

import numpy as np
from numpy.testing import assert_allclose
import pytest

from contourpy import LineType, ZInterp, contour_generator

if TYPE_CHECKING:
    import contourpy._contourpy as cpy


@pytest.fixture
def xyz_log() -> tuple[cpy.CoordinateArray, ...]:
    n = 4
    angle = 0.4
    x, y = np.meshgrid(np.linspace(0.0, 1.0, n), np.linspace(0.0, 1.0, n))

    # Rotate grid
    rot = [[np.cos(angle), np.sin(angle)], [-np.sin(angle), np.cos(angle)]]
    x, y = np.einsum("ji,mni->jmn", rot, np.dstack([x, y]))

    z = 10.0**(2.5*y)
    return x, y, z


@pytest.mark.parametrize("name", ["serial", "threaded"])
@pytest.mark.parametrize("quad_as_tri", [False, True])
def test_z_interp_log(
    xyz_log: tuple[cpy.CoordinateArray, ...],
    name: str,
    quad_as_tri: bool,
) -> None:
    x, y, z = xyz_log
    cont_gen = contour_generator(
        x, y, z, name=name, z_interp=ZInterp.Log, line_type=LineType.Separate,
        quad_as_tri=quad_as_tri)
    levels = [0.3, 1, 3, 10, 30, 100]
    all_lines = []
    for level in levels:
        expected_y = np.log10(level) / 2.5
        lines = cont_gen.lines(level)
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_Separate, lines)
        assert len(lines) == 1
        line_y = lines[0][:, 1]
        assert_allclose(line_y, expected_y, atol=1e-15)
        all_lines.append(lines)
    # The following should all produce the same lines:
    #   contour_generator(...,       z , z_interp=ZInterp.Log   ).lines(      level )
    #   contour_generator(...,   log(z), z_interp=ZInterp.Linear).lines(  log(level))
    #   contour_generator(...,  log2(z), z_interp=ZInterp.Linear).lines( log2(level))
    #   contour_generator(..., log10(z), z_interp=ZInterp.Linear).lines(log10(level))
    funcs: tuple[Callable[[Any], Any], ...] = (np.log, np.log2, np.log10)
    for func in funcs:
        cont_gen = contour_generator(
            x, y, func(z), name=name, z_interp=ZInterp.Linear, line_type=LineType.Separate,
            quad_as_tri=quad_as_tri)
        for i, level in enumerate(levels):
            lines = cont_gen.lines(func(level))
            if TYPE_CHECKING:
                lines = cast(cpy.LineReturn_Separate, lines)
            assert len(lines) == 1
            assert_allclose(lines[0], all_lines[i][0], atol=1e-15)


@pytest.mark.parametrize("name", ["serial", "threaded"])
def test_z_interp_log_saddle(name: str) -> None:
    x = y = np.asarray([-1.0, 1.0])
    z = np.asarray([[1.0, 100.0], [100.0, 1.0]])
    # z at middle of saddle quad is 10.0 for log interpolation.  Contour lines above z=10 should
    # rotate clockwise around the middle, contour lines below z=10 rotate anticlockwise.
    cont_gen = contour_generator(
        x, y, z, name=name, z_interp=ZInterp.Log, line_type=LineType.Separate)
    for level in [1.1, 9.9, 10.1, 99.9]:
        lines = cont_gen.lines(level)
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_Separate, lines)
        assert len(lines) == 2
        for line in lines:
            assert line.shape == (2, 2)
            cross_product = np.cross(line[0], line[1])
            if level > 10.0:
                assert cross_product < 0.0
            else:
                assert cross_product > 0.0


@pytest.mark.parametrize("name", ["serial", "threaded"])
def test_z_interp_negative(name: str) -> None:
    msg = "z values must be positive if using ZInterp.Log"

    z = np.asarray([[1.0, 2.0], [3.0, 4.0]])
    for value in (0.0, -1.2):
        z[1, 1] = value
        with pytest.raises(ValueError, match=msg):
            _ = contour_generator(z=z, name=name, z_interp=ZInterp.Log)

    # Mask out negative value so no exception.
    z = np.ma.masked_less_equal(z, 0.0)  # type: ignore[no-untyped-call]
    _ = contour_generator(z=z, name=name, z_interp=ZInterp.Log)



# Test functions in chunk.py

from __future__ import annotations

import pytest

from contourpy.chunk import calc_chunk_sizes, two_factors


@pytest.mark.parametrize(
    "n, factors",
    [[1, (1, 1)], [2, (2, 1)], [3, (3, 1)], [4, (2, 2)], [5, (5, 1)], [6, (3, 2)], [7, (7, 1)],
     [8, (4, 2)], [9, (3, 3)], [10, (5, 2)], [11, (11, 1)], [12, (4, 3)], [20, (5, 4)],
     [24, (6, 4)], [30, (6, 5)], [99, (11, 9)], [100, (10, 10)], [101, (101, 1)]]
)
def test_two_factors(n: int, factors: tuple[int, int]) -> None:
    assert two_factors(n) == factors


@pytest.mark.parametrize("n", [-1, -9])
def test_two_factors_invalid(n: int) -> None:
    with pytest.raises(ValueError, match=f"two_factors expects positive integer not {n}"):
        two_factors(n)


@pytest.mark.parametrize("chunk_size", [0, 1, 2, 3, 6, 9])
def test_chunk_size_1d(chunk_size: int) -> None:
    assert calc_chunk_sizes(
        chunk_size=chunk_size, chunk_count=None, total_chunk_count=None, ny=5, nx=7,
    ) == (chunk_size, chunk_size)


@pytest.mark.parametrize("x_chunk_size, y_chunk_size", [(0, 1), (1, 3), (4, 2)])
def test_chunk_size_2d(x_chunk_size: int, y_chunk_size: int) -> None:
    assert calc_chunk_sizes(
        chunk_size=(x_chunk_size, y_chunk_size), chunk_count=None, total_chunk_count=None,
        ny=5, nx=7,
    ) == (x_chunk_size, y_chunk_size)


@pytest.mark.parametrize(
    "chunk_count, res",
    [[0, (4, 6)], [1, (4, 6)], [2, (2, 3)], [3, (2, 2)], [4, (1, 2)], [5, (1, 2)], [6, (1, 1)]],
)
def test_chunk_count_1d(chunk_count: int, res: tuple[int, int]) -> None:
    assert calc_chunk_sizes(
        chunk_size=None, chunk_count=chunk_count, total_chunk_count=None, ny=5, nx=7,
    ) == res

    # Swap x and y
    assert calc_chunk_sizes(
        chunk_size=None, chunk_count=chunk_count, total_chunk_count=None, ny=7, nx=5,
    ) == res[::-1]


@pytest.mark.parametrize(
    "chunk_count, res",
    [[(0, 0), (4, 6)], [(1, 1), (4, 6)], [(2, 2), (2, 3)], [(2, 3), (2, 2)], [(9, 9), (1, 1)]  ]
)
def test_chunk_count_2d(chunk_count: tuple[int, int], res: tuple[int, int]) -> None:
    assert calc_chunk_sizes(
        chunk_size=None, chunk_count=chunk_count, total_chunk_count=None, ny=5, nx=7,
    ) == res


@pytest.mark.parametrize(
    "total_chunk_count, res",
    [[-1, (0, 0)], [0, (0, 0)], [1, (0, 0)], [2, (4, 3)], [3, (4, 2)], [4, (2, 3)]],
)
def test_total_chunk_count(total_chunk_count: int, res: tuple[int, int]) -> None:
    assert calc_chunk_sizes(
        chunk_size=None, chunk_count=None, total_chunk_count=total_chunk_count, ny=5, nx=7,
    ) == res

    # Swap x and y..
    assert calc_chunk_sizes(
        chunk_size=None, chunk_count=None, total_chunk_count=total_chunk_count, ny=7, nx=5,
    ) == res[::-1]


def test_calc_chunk_sizes_invalid() -> None:
    msg = "Only one of chunk_size, chunk_count and total_chunk_count should be set"
    with pytest.raises(ValueError, match=msg):
        calc_chunk_sizes(chunk_size=1, chunk_count=1, total_chunk_count=None, ny=2, nx=2)
    with pytest.raises(ValueError, match=msg):
        calc_chunk_sizes(chunk_size=1, chunk_count=None, total_chunk_count=1, ny=2, nx=2)
    with pytest.raises(ValueError, match=msg):
        calc_chunk_sizes(chunk_size=None, chunk_count=1, total_chunk_count=1, ny=2, nx=2)
    with pytest.raises(ValueError, match=msg):
        calc_chunk_sizes(chunk_size=1, chunk_count=1, total_chunk_count=1, ny=2, nx=2)

    msg = r"\(ny, nx\) must be at least \(2, 2\)"
    with pytest.raises(ValueError, match=msg):
        calc_chunk_sizes(chunk_size=1, chunk_count=None, total_chunk_count=None, ny=1, nx=2)
    with pytest.raises(ValueError, match=msg):
        calc_chunk_sizes(chunk_size=1, chunk_count=None, total_chunk_count=None, ny=2, nx=1)
    with pytest.raises(ValueError, match=msg):
        calc_chunk_sizes(chunk_size=1, chunk_count=None, total_chunk_count=None, ny=1, nx=1)



from __future__ import annotations

from typing import Any, Sequence

from docutils import nodes


class Table:
    _ncols: int
    _table: Any
    _tbody: Any
    _tgroup: Any

    def __init__(self, ncols: int) -> None:
        self._ncols = ncols
        self._table = nodes.table()
        self._tgroup = nodes.tgroup(cols=ncols)
        for _ in range(ncols):
            colspec = nodes.colspec(colwidth=1)
            self._tgroup.append(colspec)
        self._table += self._tgroup
        self._tbody = None

    def add_header(self, items: Sequence[Any]) -> None:
        if len(items) != self._ncols:
            raise RuntimeError(f"Expect {self._ncols} header items but {len(items)} specified")

        thead = nodes.thead()
        self._tgroup += thead

        row = nodes.row()
        thead += row

        for item in items:
            entry = nodes.entry()
            row += entry
            entry += nodes.paragraph(text=item)

    def add_row(self, items: Sequence[Any]) -> None:
        if len(items) != self._ncols:
            raise RuntimeError(f"Expect {self._ncols} row items but {len(items)} specified")

        if self._tbody is None:
            self._tbody = nodes.tbody()
            self._tgroup += self._tbody

        row = nodes.row()
        self._tbody += row

        for item in items:
            entry = nodes.entry()
            row += entry
            entry += nodes.paragraph(text=item)

    def get(self) -> Any:
        return self._table



from __future__ import annotations

from contourpy import LineType, contour_generator
from contourpy.util.mpl_renderer import MplTestRenderer

from .bench_base import BenchBase
from .util_bench import corner_mask_to_bool, corner_masks, datasets, problem_sizes


class BenchLinesMpl20xxRender(BenchBase):
    params: tuple[list[str], list[str], list[LineType], list[str | bool], list[int]] = (
        ["mpl2005", "mpl2014"], datasets(), [LineType.SeparateCode], corner_masks(),
        problem_sizes(),
    )
    param_names: tuple[str, ...] = ("name", "dataset", "line_type", "corner_mask", "n")

    def setup(
        self, name: str, dataset: str, line_type: LineType, corner_mask: str | bool, n: int,
    ) -> None:
        self.set_xyz_and_levels(dataset, n, corner_mask != "no mask")

    def time_lines_mpl20xx_render(
        self, name: str, dataset: str, line_type: LineType, corner_mask: str | bool, n: int,
    ) -> None:
        if name == "mpl2005" and corner_mask is True:
            raise NotImplementedError  # Does not support corner_mask=True
        cont_gen = contour_generator(
            self.x, self.y, self.z, name=name, line_type=line_type,
            corner_mask=corner_mask_to_bool(corner_mask),
        )
        renderer = MplTestRenderer()
        for i, level in enumerate(self.levels):
            lines = cont_gen.lines(level)
            renderer.lines(lines, line_type, color=f"C{i}")
        renderer.save(f"lines_{name}_{corner_mask}_{line_type}_{n}.png")
